// This is a planner that calculates the collision coordinates of a vehicle
// The following information must be provided:
// - collision coordinates (cx, cy)
// - vehicle startcoordinates (sx, sy)
// - vehiclepart that collides (front, rear etc...)
// - vehiclesize; width, length

package  se.lth.cs.carsimgui;


class CollisionPlanner implements Planner {
   
   // Public properties
   
   // Collision coordinates
   public float collX = 0.0f;
   public float collY = 0.0f;
   
   // Vehiclepart
   public int vehiclepart = Constant.LEFTSIDE;
   
   // Vehiclesize
   public float vehicleLength = 3.55f;
   public float vehicleWidth = 1.25f;
   
   // Private members
   Event collEvent;
   private boolean errorsDetected = false;
   
   // Constructor
   public CollisionPlanner(Event e) {
      collEvent = e;
   }
   
   // Start the calculation of the collision trajectory
   public void plan() {
      if (vehiclepart == Constant.LEFTSIDE || vehiclepart == Constant.RIGHTSIDE)
         sideCollision();
      if (vehiclepart == Constant.FRONT || vehiclepart == Constant.REAR)
         frontrearCollision();
   }
   
   // This method calculates the coordinates of a side collision.
   private void sideCollision() {
      double vx = 0.0d;
      double vy = 0.0d;
      
      // Define the vehicle startcoordinates
      double sx = (double)collEvent.startPosX;
      double sy = (double)collEvent.startPosY;
      
      // Define the collision coordinates
      double cx = (double)collX;
      double cy = (double)collY;
      
      // With a side collision, the circle radius is half the width of the vehicle
      double r = (double)vehicleWidth/2.0d;
      
      // Calculate the distance between the collision point and the car startpoint
      double dist = Math.sqrt(Math.pow(cx-sx,2) + Math.pow(cy-sy,2));
      
      // Calculate the alpha angle between the two possible sidecollision points and the x-axis
      double alphaI = Math.acos(r/dist);
      double alphaII = -Math.acos(r/dist);
      
      // Calculate the turning angle for correction:
      double deltaX = sx - cx;
      double deltaY = sy - cy;
      double turnAngle = 0.0d;
      if (deltaX == 0) {
         if (cy > sy)
            turnAngle = -Math.PI/2.0d;
         else
            turnAngle = Math.PI/2.0d;
      } else
         turnAngle = Math.atan(deltaY/deltaX);
      if (deltaX < 0)
         turnAngle = turnAngle + Math.PI;
      
      // Decide which alpha to use 
      double finalAlpha = 0.0d;
      if (vehiclepart == Constant.RIGHTSIDE)
         finalAlpha = alphaI;
      else
         finalAlpha = alphaII;
      
      // Calculate the position of the collision point with rotation correction and translation correction
      vx = r*Math.cos(finalAlpha + turnAngle) + cx;
      vy = r*Math.sin(finalAlpha + turnAngle) + cy; 
      
      // Set the Event endpoint
      collEvent.endPosX = (float)vx;
      collEvent.endPosY = (float)vy;      
   }
   
   // This method calculates the coordinates of a frontal or rear collision.
   private void frontrearCollision() {
      double vx = 0.0d;
      double vy = 0.0d;
      
      // Define the vehicle startcoordinates
      double sx = (double)collEvent.startPosX;
      double sy = (double)collEvent.startPosY;
      
      // Define the collision coordinates
      double cx = (double)collX;
      double cy = (double)collY;
      
      // if it is a frontal collision, the initAngle = 0, otherwise it is pi
      double initAngle = 0.0d;
      if (vehiclepart == Constant.REAR)
         initAngle = Math.PI;
      
      // Calculate the turning angle for correction:
      double deltaX = sx - cx;
      double deltaY = sy - cy;
      double turnAngle = 0.0d;
      if (deltaX == 0) {
         if (cy > sy)
            turnAngle = -Math.PI/2.0d;
         else
            turnAngle = Math.PI/2.0d;
      } else
         turnAngle = Math.atan(deltaY/deltaX);
      if (deltaX < 0)
         turnAngle = turnAngle + Math.PI;
      
      // The radius is half the length of the car.
      double r = (double)vehicleLength/2.0d;
            
      // Calculate the position of the collision point with rotation correction and translation correction
      vx = r*Math.cos(initAngle + turnAngle) + cx;
      vy = r*Math.sin(initAngle + turnAngle) + cy; 
      
      // Set the Event endpoint
      collEvent.endPosX = (float)vx;
      collEvent.endPosY = (float)vy;  
   }
 
   // Returns true if there are errors detected.  
   public boolean errors() {
      return errorsDetected;
   }
}