// This planner tries to fill in unknown values in the accident info object

// 	Name: PrePlanner (class)
//	Author: Arjan Egges
//	Version: 1.0

//	Version 1.1 by Bastian Schulz
//	Extensions:
//		transforms Constant.LEAVEROAD to ..._LEFT resp. ..._RIGHT
//		transforms Constant.OVERTURN  to ..._LEFT resp. ..._RIGHT
//		sets overturn attribute (of actor or vistim in Collision) from unknown to left or right

package  se.lth.cs.carsimgui;

class PrePlanner implements Planner {
   
   // private members
   private AccidentInfo ainfo;
   
   public PrePlanner(AccidentInfo ai) {
      ainfo = ai;
   }
   
   public void plan() {
      if (ainfo.getCollisionInfoAmount() == 0) {
         Messenger.message("No accident modeling required.\n");
         fillInUnknownDirections();
         return;
      }
      fillInUnknownCollisionValues();
   }
   
   private void fillInUnknownCollisionValues() {

      // Plan the first accident
      CollisionInfo cinfo = ainfo.getCollisionInfo(0);
      DynamicInfo actorInfo = ainfo.getDynamicInfo(cinfo.actorIndex);
      DynamicInfo victimInfo = ainfo.getDynamicInfo(cinfo.victimIndex);
      
      // If the direction of the victim and the part of the victim are unknown
      if (cinfo.victimPart == Constant.UNKNOWN_PART && victimInfo.initDirection == DynamicInfo.UNKNOWN_DIRECTION) {
         victimInfo.initDirection = DynamicInfo.NORTH;
         cinfo.victimPart = Constant.LEFTSIDE;

      // If only the direction is unknown
      } else if (victimInfo.initDirection == DynamicInfo.UNKNOWN_DIRECTION) {
         if (cinfo.victimPart == Constant.FRONT)
            victimInfo.initDirection = DynamicInfo.WEST;
         else if (cinfo.victimPart == Constant.LEFTSIDE)
            victimInfo.initDirection = DynamicInfo.NORTH;
         else if (cinfo.victimPart == Constant.RIGHTSIDE)
            victimInfo.initDirection = DynamicInfo.SOUTH;
         else if (cinfo.victimPart == Constant.REAR)
            victimInfo.initDirection = DynamicInfo.NORTH;

      // If only the victim part is unknown
      } else if (cinfo.victimPart == Constant.UNKNOWN_PART) {
         if (victimInfo.initDirection == DynamicInfo.WEST)
            cinfo.victimPart = Constant.LEFTSIDE;
         else if (victimInfo.initDirection == DynamicInfo.NORTH)
            cinfo.victimPart = Constant.LEFTSIDE;
         else if (victimInfo.initDirection == DynamicInfo.SOUTH)
            cinfo.victimPart = Constant.RIGHTSIDE;
         else if (victimInfo.initDirection == DynamicInfo.EAST)
            cinfo.victimPart = Constant.REAR;
      }
      
      fillInUnknownDirections();

      // Plan the rest of the unknown values for the parts
      for (int i=0; i<ainfo.getCollisionInfoAmount(); i++) {
         if (ainfo.getCollisionInfo(i).actorPart == Constant.UNKNOWN_PART)
            ainfo.getCollisionInfo(i).actorPart = Constant.FRONT;
         if (ainfo.getCollisionInfo(i).victimPart == Constant.UNKNOWN_PART)
            ainfo.getCollisionInfo(i).victimPart = Constant.FRONT;
      }
      
      // set <actor overturn="unknown" ...> to overturn="left" resp. overturn="right"
	for (int i=0; i<ainfo.getCollisionInfoAmount(); i++)
	{
		CollisionInfo ci = ainfo.getCollisionInfo(i);
		
		if (ci.actorOverturn == Constant.OVERTURN)
		{
			if (ci.actorPart == Constant.FRONT)
				ci.actorOverturn = Constant.OVERTURN_RIGHT; // could be LEFT as well
			else if (ci.actorPart == Constant.REAR)
				ci.actorOverturn = Constant.OVERTURN_LEFT; // could be RIGHT as well
			else if (ci.actorPart == Constant.LEFTSIDE)
				ci.actorOverturn = Constant.OVERTURN_RIGHT;
			else if (ci.actorPart == Constant.RIGHTSIDE)
				ci.actorOverturn = Constant.OVERTURN_LEFT;
		}
		
		if (ci.victimOverturn == Constant.OVERTURN)
		{
			if (ci.victimPart == Constant.FRONT)
				ci.victimOverturn = Constant.OVERTURN_RIGHT; // could be LEFT as well
			else if (ci.victimPart == Constant.REAR)
				ci.victimOverturn = Constant.OVERTURN_LEFT; // could be RIGHT as well
			else if (ci.victimPart == Constant.LEFTSIDE)
				ci.victimOverturn = Constant.OVERTURN_RIGHT;
			else if (ci.victimPart == Constant.RIGHTSIDE)
				ci.victimOverturn = Constant.OVERTURN_LEFT;
		}
	}
   }
   
   public void fillInUnknownDirections() {
	// The first actor direction is east; if not known
	if (ainfo.getDynamicInfo(ainfo.getCollisionInfo(0).actorIndex).initDirection == DynamicInfo.UNKNOWN_DIRECTION)
		ainfo.getDynamicInfo(ainfo.getCollisionInfo(0).actorIndex).initDirection = DynamicInfo.EAST;
         
	// Fill in all unknown directions
	for (int i=0; i<ainfo.getDynamicAmount(); i++)
	{
		if (ainfo.getDynamicInfo(i).initDirection == DynamicInfo.UNKNOWN_DIRECTION)
			ainfo.getDynamicInfo(i).initDirection = DynamicInfo.EAST;
	}
      
	// transform LEAVEROAD to the _LEFT or _RIGHT option
	boolean left = true;
	
	for (int i=0; i<ainfo.getDynInfoAmount(); i++)
	{
		EventChain evc = ainfo.getDynamicInfo(i).ceb;
		int eventkind;
		for (int j=0; j < evc.getEventAmount(); j++)
		{
			eventkind = evc.getEvent(j).kind;
			if (eventkind == Constant.STRAIGHTROAD)
				left = false;
			else if (eventkind == Constant.TURNLEFT)
				left = false;
			else if (eventkind == Constant.TURNRIGHT)
				left = true;
			else if (eventkind == Constant.STOP)
				left = false;
			else if (eventkind == Constant.CHANGELANE_LEFT)
				left = true;
			else if (eventkind == Constant.CHANGELANE_RIGHT)
				left = false;
			else if (eventkind == Constant.LEAVEROAD)
				if (left)
					evc.getEvent(j).kind = Constant.LEAVEROAD_LEFT;
				else
					evc.getEvent(j).kind = Constant.LEAVEROAD_RIGHT;
		}
	}

	// transform OVERTURN to the _LEFT or _RIGHT option
	left = true;
	
	for (int i=0; i<ainfo.getDynInfoAmount(); i++)
	{
		EventChain evc = ainfo.getDynamicInfo(i).ceb;
		int eventkind;
		for (int j=0; j < evc.getEventAmount(); j++)
		{
			eventkind = evc.getEvent(j).kind;
			if (eventkind == Constant.STRAIGHTROAD)
				left = true;
			else if (eventkind == Constant.TURNLEFT)
				left = false;
			else if (eventkind == Constant.TURNRIGHT)
				left = true;
			else if (eventkind == Constant.CHANGELANE_LEFT)
				left = true;
			else if (eventkind == Constant.CHANGELANE_RIGHT)
				left = false;
			else if (eventkind == Constant.LEAVEROAD_LEFT)
				left = false;
			else if (eventkind == Constant.LEAVEROAD_RIGHT)
				left = true;
			else if (eventkind == Constant.OVERTURN)
				if (left)
					evc.getEvent(j).kind = Constant.OVERTURN_LEFT;
				else
					evc.getEvent(j).kind = Constant.OVERTURN_RIGHT;
		}
	}
   }
   
   public boolean errors() {
      return false;
   }
}