// This planner plan the temporal values of the trajectory

// 	Name: TemporalPlanner (class)
//	Author: Arjan Egges
//	Version: 1.0


package  se.lth.cs.carsimgui;

class TemporalPlanner implements Planner {

	// Planning properties
	public float stopPerc = 15.0f;
	public float accPerc = 20.0f;
	
	// Private members
 	private AccidentInfo ainfo;
	private boolean errorsDetected;
	private float errorRange = 0.0001f;
	
	// Synchronisation point array
	private float[] spArr;
	
	// Constructor
	public TemporalPlanner(AccidentInfo ai)
	{
		ainfo = ai;
		errorsDetected = false;
		
		// Set the general planning properties
		stopPerc = Messenger.properties.stopPerc;
		
		if (ainfo.getCollisionInfoAmount() > 0)
		{
			// Initialise the synchronisation point array
			spArr = new float[ainfo.getCollisionInfoAmount()];
			// Calculate the synchronisation values
			spArr[0] = 1.0f - accPerc/100.0f;
			int i = 1;
			while (i < ainfo.getCollisionInfoAmount())
			{
				spArr[i] = spArr[i-1] + accPerc/100.0f/(ainfo.getCollisionInfoAmount()-1);
				i++;
			}
		}
		else
		{
			spArr = new float[1];
			spArr[0] = 1.0f - accPerc/100.0f;
			
			// spArr[0] = 1.0f; why don't use the whole time 0.0 .. 1.0 if there are no collisions at all?
			// Because: from 0.8 to 1.0 are only collisions. Even when there is only one collision,
			// the collisions is at the timestamp 0.8 (and not 1.0). To keep the same driving speed
			// scenes without any collisions have to stop at 0.8 as well.
		}
	}

	// Planning method
	public void plan()
	{
		// The planning consists of calculating the percentages of every trajectory segment
		for (int index=0; index < ainfo.getDynamicAmount(); index++)
		{
			calculateTimeStamps(index);
		}
		
		// normalize
		// find the greatest percentage value that is greater than 1.0f
		// if there is one, divide ALL percentage values by this value
		
		float maxPercentage = 1.0f;
		Trajectory t;
		for (int i=0; i < ainfo.getDynamicAmount(); i++)
		{
			t = ainfo.getDynamicObject(i).getTrajectory();
			if (t.get(t.getAmount()-1).percentage > maxPercentage)
				maxPercentage = t.get(t.getAmount()-1).percentage;
		}
		if (maxPercentage > 1)
		{
			for (int i=0; i < ainfo.getDynamicAmount(); i++)
			{
				t = ainfo.getDynamicObject(i).getTrajectory();
				for (int j=0; j < t.getAmount(); j++)
				{
					t.get(j).percentage = t.get(j).percentage/maxPercentage;
				}
			}
		}
		
		// If the total percentage != 1.0, add a stop
		for (int i=0; i < ainfo.getDynamicAmount(); i++)
		{
			t = ainfo.getDynamicObject(i).getTrajectory();
			if (Math.abs( t.get(t.getAmount()-1).percentage - 1.0f) >= errorRange )
      			{
      				TrajectoryPoint tp = new TrajectoryPoint(t.get(t.getAmount()-1));
      				tp.percentage = 1.0f;
      				t.add(tp);
      			}
		}
	}
  
	// Calculates the time stamps (percentages) of a trajectory t.
	private void calculateTimeStamps(int trajNr)
	{
		Trajectory t = ainfo.getDynamicObject(trajNr).getTrajectory();
		int tpAmount = t.getAmount();
		if (tpAmount == 0)
			return;
		
		int fromTP = 0; // from trajectory point
		int   toTP = 0; //   to trajectory point
		float percentage_from_to      = 0.0f;
		float percentage_from         = 0.0f;
		float percentage_to           = 0.0f;
		float tempPercentage          = 0.0f;
		float distance                = 0.0f;
		float distance_from_to        = 0.0f;
		float distance_per_percentage = 0.0f; // needed for overturn after the last collision
		// then no percentage value is available, so the last speed is kept
		
		int stopCount = 0;  // number of STOP events of the whole trajectory
		int stopFromToCount = 0;  // number of STOP events within a from..to-segment
		float stopFactor = 0; // is assigned to the sum of all stops during a from..to-segment 
		float restFactor = 0; // is assigned to the sum of all other events during a from..to-segment
		
		for (int index=1; index < t.getAmount(); index++) {
			if (distance(t.get(index-1), t.get(index)) == 0.0f)
				stopCount++;
		}
		
		stopFactor = stopPerc / 100.0f;
		
		t.get(0).percentage = 0.0f;
		
		while( toTP + 1 < tpAmount )
		{
			stopFromToCount = 0;
			fromTP = toTP;
			toTP++;
			distance_from_to = distance(t.get(fromTP), t.get(toTP));
			
			while((toTP + 1< tpAmount) && (t.get(toTP).collisionIndex == -1))
			{
				distance = distance(t.get(toTP), t.get(toTP+1));
				
				
				if (distance == 0.0f)
				{
					stopFromToCount++;
				}
				else
				{
					distance_from_to += distance;
				}
				toTP++;
			}
			
			
			if (t.get(fromTP).collisionIndex != -1)
				percentage_from = spArr[t.get(fromTP).collisionIndex];
				
			if (t.get(toTP).collisionIndex != -1) // a collision was found
			{
				percentage_to   = spArr[t.get(toTP).collisionIndex];
			}
			else                                  // the end of the trajectory is reached
			{
				if(fromTP == 0.0f) // no collisions in the whole trajectory
				{
					percentage_to = spArr[0];
				}
				else
				{
					percentage_to = percentage_from + distance_from_to / distance_per_percentage;
				}
			}
			
			percentage_from_to = percentage_to - percentage_from;
			tempPercentage = percentage_from;
			distance_per_percentage = distance_from_to / percentage_from_to;
			
			
			//stopFactor = percentage_from_to * stopPerc / 100.0f;
			if (stopCount>0)
				restFactor = percentage_from_to - stopFromToCount*stopFactor/stopCount;
			else
				restFactor = percentage_from_to;
			
			for(int i = fromTP+1; i <= toTP; i++)
			{
				distance = distance(t.get(i-1), t.get(i));
				if (distance == 0.0f)
					tempPercentage += stopFactor/(float)stopCount;
				else
					tempPercentage += distance / distance_from_to * restFactor;
				t.get(i).percentage = tempPercentage;
			}
		}
	}


	// Calculates the distance between two trajectory points p1, p2
	private float distance(TrajectoryPoint p1, TrajectoryPoint p2)
	{
		double deltaX = (double)(p2.x - p1.x);
		double deltaY = (double)(p2.y - p1.y);
		float d = (float)Math.sqrt(Math.pow(deltaX,2) + Math.pow(deltaY,2));
		return d;
	}
	
	// Returns true if there are errors detected
	public boolean errors() {
		return errorsDetected;
	}
}
